#include <stdlib.h>
#include "geosim.h"
#include "gspriv.h"
#include "ietypes.h"
#include "iemisc.h"

extern IEIconfig iestate;
extern GSColor blackCol;

#define MAX_CLIP_LEV 2
#define CENTER /* center text vertically in field */

static int clx[MAX_CLIP_LEV+1] = {0}, 
			cly[MAX_CLIP_LEV+1] = {0}, 
			clwd[MAX_CLIP_LEV+1] = {SCREENWIDTH}, 
			clht[MAX_CLIP_LEV+1] = {SCREENHEIGHT};
static int clipLev = 0;

#define NUM_RESERVED_WORDS 21

static char *gReserved[NUM_RESERVED_WORDS] = { /* +++ init from a file? */
	"alias", "button", "buttons", "buttonstuff", "cancel",
	"colorlist", "dragarea", "highlight", "field", "function",
	"highlight", "label", "labels", "line", "msg", "palette",
	"plabel", "popup", "rect", "redrawfunc", "window"
	};

void ShadowRect(int x, int y, int wd, int ht, int th, ConcavityType ls)
{
	x-=th; y-=th; wd+=th; ht+=th;
	GSsetlinesize(th);
	GSsetcolor(GSconvertcolor((ls==Concave)?"black":"dgray"));
	GSwdrawline(x, y, x + wd, y);
	GSwdrawline(x, y, x, y + ht);
	GSsetcolor(GSconvertcolor((ls!=Concave)?"black":"dgray"));
	GSwdrawline(x + wd, y, x + wd, y + ht);
	GSwdrawline(x, y + ht, x + wd, y + ht);
}

void DeleteWindow(IEWindow *w)
{
	if (iestate.winlisthead == w)
		iestate.winlisthead = w->nextwin;
	if (iestate.winlisttail == w)
		iestate.winlisttail = w->prevwin;
	
	if (w->prevwin)
		w->prevwin->nextwin = w->nextwin;
	if (w->nextwin)
		w->nextwin->prevwin = w->prevwin;
	
	free(w);
}

void DrawNumField(char *fName, int val)
{
	BigString t;
	sprintf(t, "%d", val);
	DrawTextField(fName, t);
}

void DrawTextField(char *fName, char *text)
{
	int x, y, wd, ht;
#ifdef CENTER
	int tht;
#endif

	GSColor col;
		
	GSgetfieldrect(fName, &x, &y, &wd, &ht, &col);
	
	GSsetcolor(col);
	GSwfillrect(x, y, wd, ht);
	
	ShadowRect(x, y, wd, ht, 1, Concave);
	
	GSsetcolor(blackCol);
	GSsetfont(SMALL);
	
#ifdef CENTER
	tht = GStextheight(text);
	GSwwritemsg(x, y+ht-((ht-tht)/2), text);
#else
	GSwwritemsg(x, y+SMALL, text);
#endif
}

void DeleteButton(IEWindow *w, IEButton *b)
{
	if (w->firstbutton == b)
		w->firstbutton = b->nextbutton;
	if (w->lastbutton == b)
		w->lastbutton = b->prevbutton;
		
	if (b->prevbutton)
		b->prevbutton->nextbutton = b->nextbutton;
	if (b->nextbutton)
		b->nextbutton->prevbutton = b->prevbutton;
	
	free(b);
}

void DeleteDragArea(IEWindow *w, IEDragArea *b)
{
	if (w->firstdragarea == b)
		w->firstdragarea = b->nextdragarea;
	if (w->lastdragarea == b)
		w->lastdragarea = b->prevdragarea;
		
	if (b->prevdragarea)
		b->prevdragarea->nextdragarea = b->nextdragarea;
	if (b->nextdragarea)
		b->nextdragarea->prevdragarea = b->prevdragarea;
	
	free(b);
}

void DeleteField(IEWindow *w, IEField *b)
{
	if (w->firstfield == b)
		w->firstfield = b->nextfield;
	if (w->lastfield == b)
		w->lastfield = b->prevfield;
		
	if (b->prevfield)
		b->prevfield->nextfield = b->nextfield;
	if (b->nextfield)
		b->nextfield->prevfield = b->prevfield;
		
	free(b);
}

void DeleteGraphic(IEWindow *w, IEGraphic *b)
{
	if (w->firstgraphic == b)
		w->firstgraphic = b->nextgraphic;
	if (w->lastgraphic == b)
		w->lastgraphic = b->prevgraphic;
		
	if (b->prevgraphic)
		b->prevgraphic->nextgraphic = b->nextgraphic;
	if (b->nextgraphic)
		b->nextgraphic->prevgraphic = b->prevgraphic;
	
	free(b);
}

void DeletePopup(IEPopup *p)
{	
	if (iestate.popuplisthead == p)
		iestate.popuplisthead = p->nextpopup;
	if (iestate.popuplisttail == p)
		iestate.popuplisttail = p->prevpopup;
		
	if (p->prevpopup)
		p->prevpopup->nextpopup = p->nextpopup;
	if (p->nextpopup)
		p->nextpopup->prevpopup = p->prevpopup;

	free(p);
}

IEPopup *GetPopupNum(int pn) /* start with 0 */
{
	IEPopup *p = iestate.popuplisthead;
	
	if (pn < 0)
		return NULL;
		
	while (p != NULL) {
		if (pn == 0)
			return p;
		pn--;
		p = p->nextpopup;
		}
	
	return NULL;
}

IEMenuItem *GetPlabelNum(int n, IEPopup *p)
{
	IEMenuItem *m = p->itemlisthead;
	
	if (n < 0)
		return NULL;
		
	while (m != NULL) {
		if (n==0)
			return m;
		n--;
		m = m->nextmenuitem;
		}
		
	return NULL;
}

void SetPopupDims(IEPopup *p)
{
	int len;
	IEMenuItem *m = p->itemlisthead;
	FontSizeType oldfont = GSPgetfont(); 
	
	p->wd = 2*POPUPOFFSET;
	p->ht = TSIZE/2;
	/* save font */
	GSsetfont(LARGE);
	while (m != NULL) {
		len = GStextwidth(m->text) + 2*POPUPOFFSET;
		if (p->checklist)
			len += CHECKWID;
		if (len > p->wd)
			p->wd = len;
		if (m->active != INVISIBLE)
			p->ht += TSIZE;

		m = m->nextmenuitem;
		}
	
	GSsetfont(oldfont);
}

boolean point_in_rect(int ptx, int pty,
				int rx, int ry, int rwd, int rht)
{
	if ((ptx < rx) || (pty < ry))
		return FALSE;
	if ((ptx > (rx+rwd)) || (pty > (ry + rht)))
		return FALSE;
	return TRUE;
}

#define EPS 8

boolean point_hits_line(int ptx, int pty,
				int l1x, int l1y, int l2x, int l2y)
{
	long len_sq;
	long x_rot, y_rot;
	
	l2x -= l1x;
	ptx -= l1x;
	l2y -= l1y;
	pty -= l1y;
	
	len_sq = l2x*l2x + l2y*l2y;
	
	if (len_sq == 0)
		return FALSE;
	
	x_rot = l2x*ptx + l2y*pty;
	
	if ((x_rot < 0) || (x_rot > len_sq))
		return FALSE;
		
	y_rot = labs(l2x*pty - l2y*ptx);
			
	if ( y_rot*y_rot > (EPS*EPS)*len_sq )
/*	if ( y_rot > EPS*((long)sqrt((double)len_sq)) )*/
		return FALSE;
	else
		return TRUE;
}

boolean r_intersect(int r1x, int r1y, int r1wd, int r1ht,
					int r2x, int r2y, int r2wd, int r2ht)
{
	if ((r2x+r2wd < r1x) || (r2x > r1x+r1wd))
		return FALSE;
		
	if ((r2y+r2ht < r1y) || (r2y > r1y+r1ht))
		return FALSE;
		
	return TRUE;
}

void calc_union_rect(int r1x, int r1y, int r1wd, int r1ht,
					int r2x, int r2y, int r2wd, int r2ht,
					int *cx, int *cy, int *cwd, int *cht)
{
	int tempx, tempy;
	*cx = MIN(r1x, r2x);
	*cy = MIN(r1y, r2y);
	tempx = MAX(r1x+r1wd, r2x+r2wd);
	tempy = MAX(r1y+r1ht, r2y+r2ht);
	*cwd = tempx - *cx;
	*cht = tempy - *cy;
}

void calc_sect_rect(int r1x, int r1y, int r1wd, int r1ht,
					int r2x, int r2y, int r2wd, int r2ht,
					int *cx, int *cy, int *cwd, int *cht)
{
	int tempx, tempy;
	*cx = MAX(r1x, r2x);
	*cy = MAX(r1y, r2y);
	tempx = MIN(r1x+r1wd, r2x+r2wd);
	tempy = MIN(r1y+r1ht, r2y+r2ht);
	*cwd = tempx - *cx;
	*cht = tempy - *cy;
}

void set_label_dims(IEGraphic *g, int x, int y)
{
	g->gr.x = x;
	GSsetfont((g->gr.type == GR_LABEL)? LARGE : SMALL);
	g->gr.wd = GStextwidth(g->gr.text);
	g->gr.ht = (g->gr.type == GR_LABEL)? LARGE : SMALL;
	g->gr.y = y - g->gr.ht;
}

void PushClip(int x, int y, int wd, int ht)
{
	if (clipLev >= MAX_CLIP_LEV)
		GSelog(-1, "PushClip overflow.\n");
	
	calc_sect_rect(clx[clipLev], cly[clipLev], 
					clwd[clipLev], clht[clipLev],
					x, y, wd, ht,
					&clx[clipLev+1], &cly[clipLev+1], 
					&clwd[clipLev+1], &clht[clipLev+1]);
			
	clipLev++;
	GSPset_clip_rect(clx[clipLev], cly[clipLev], clwd[clipLev], clht[clipLev]);
}

void PopClip()
{
	clipLev--;
	if (clipLev < 0)
		GSelog(-1, "PopClip() underflow.\n");
	
	GSPset_clip_rect(clx[clipLev], cly[clipLev], clwd[clipLev], clht[clipLev]);
}

void ResetClip()
{
	clipLev = 0;
	GSPset_clip_rect(clx[clipLev], cly[clipLev], clwd[clipLev], clht[clipLev]);
}

#if 0
/* experimental name-validation routine */

boolean VerifyName(char *name, char *newName, char *fieldName, BigString errorMsg)
{
	/* Make sure that:
		* Name is not a reserved word.  Reserved words are case-insensitive.
		* Name is not zero-length.
		
		Other validations can be implemented in this function.
	*/
	int i;
	
	if (strlen(name) == 0) {
		if (errorMsg)
			sprintf(errorMsg, "Error: %s is zero-length.", fieldName);
		return FALSE;
		}
	
	for (i=0; i< NUM_RESERVED_WORDS; i++)
		if (GSstricmp(gReserved[i], name) == 0) {
			if (errorMsg) {
				sprintf(errorMsg, "Error: %s is a reserved word.", fieldName);
				return FALSE;
				}
			}

	return TRUE;			
}

#endif
